# Generated by Django 2.2.1 on 2020-07-27 14:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0004_keyvaluestore'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='sw_version',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]
