# Generated by Django 2.2.1 on 2019-11-12 23:02

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AutoAlignment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Feature',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('visible', models.BooleanField(blank=True, default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Generator',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('params', models.TextField()),
                ('static', models.BooleanField()),
                ('visible', models.BooleanField(blank=True, default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Sequence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=128)),
                ('params', models.TextField()),
                ('sequence', models.TextField(blank=True)),
                ('generator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Generator')),
            ],
        ),
        migrations.CreateModel(
            name='Subject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name='System',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('path', models.TextField()),
                ('archive', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('visible', models.BooleanField(blank=True, default=True)),
            ],
        ),
        migrations.CreateModel(
            name='TaskEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('params', models.TextField()),
                ('report', models.TextField()),
                ('notes', models.TextField()),
                ('visible', models.BooleanField(blank=True, default=True)),
                ('backup', models.BooleanField(blank=True, default=False)),
                ('feats', models.ManyToManyField(to='tracker.Feature')),
                ('sequence', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.Sequence')),
                ('subject', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Subject')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Task')),
            ],
        ),
        migrations.AddField(
            model_name='sequence',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Task'),
        ),
        migrations.CreateModel(
            name='Decoder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=128)),
                ('path', models.TextField()),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.TaskEntry')),
            ],
        ),
        migrations.CreateModel(
            name='DataFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('local', models.BooleanField(default=True)),
                ('archived', models.BooleanField(default=False)),
                ('path', models.CharField(max_length=256)),
                ('entry', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.TaskEntry')),
                ('system', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.System')),
            ],
        ),
        migrations.CreateModel(
            name='Calibration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=128)),
                ('params', models.TextField()),
                ('subject', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.Subject')),
                ('system', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='tracker.System')),
            ],
        ),
    ]
