# Generated by Django 3.1.4 on 2021-10-17 15:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0009_auto_20210630_1019'),
    ]

    operations = [
        migrations.CreateModel(
            name='Experimenter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
            ],
        ),
    ]
