# Generated by Django 2.2.1 on 2019-11-13 21:58

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0002_auto_20191112_2306'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datafile',
            name='system',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.System'),
        ),
        migrations.AlterField(
            model_name='task',
            name='visible',
            field=models.BooleanField(blank=True, default=True),
        ),
    ]
