# Generated by Django 3.1.4 on 2021-10-17 15:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0010_experimenter'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='experimenter',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.experimenter'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='taskentry',
            name='project',
            field=models.TextField(default='testing'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='taskentry',
            name='session',
            field=models.TextField(default='testing'),
            preserve_default=False,
        ),
    ]
