# Generated by Django 3.1.4 on 2021-05-16 15:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0007_taskentry_template'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='metadata',
            field=models.TextField(default=''),
        ),
    ]
